/*
Copyright notice:
Jenny - The Java iButton password safe - copyright 2004 kamikazi / blastzone.net
This program uses some code obtained from Dallas Semiconductor.  This code is in the public domain.
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

This notice must be included with any future verions of this program.
*/
/*
DEVELOPERS NOTES

  Comments on choices made for datatypes:
  It would seem logical to collect and store the password in an object of type java.lang.String.
  However, here's the caveat: Objects of type String are immutable, i.e., there are no methods defined
  that allow you to overwrite or zero out the contents of a String after usage.
  This feature makes String objects unsuitable for storing security sensitive information
  such as user passwords. Sensitive information is therefore stored in character arrays.

  This program is PKCS#5 v2 compliant. To use Password-Based Encryption (PBE) we have to specify a
  salt and an iteration count. The same salt and iteration count that are used for encryption must
  be used for decryption.  The salt must be over 64bits, iteration count over 1000.
  The salt and iteration values need not be secret. Information on PKCS#5 can be found at:
  http://www.rsasecurity.com/rsalabs/pkcs/pkcs-5/index.html


*/
//dependencies for main & crypto classes
import java.io.*;
import java.nio.charset.*;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.util.Arrays;
import java.util.Random;
import java.util.Date;
import java.math.BigInteger;
import java.security.*;
import java.security.spec.*;
import java.security.interfaces.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import javax.crypto.interfaces.*;
import com.sun.crypto.provider.SunJCE;

//dependencies for iButton classes
import java.util.Vector;
import java.util.Enumeration;
import com.dalsemi.onewire.OneWireException;
import com.dalsemi.onewire.OneWireAccessProvider;
import com.dalsemi.onewire.adapter.DSPortAdapter;
import com.dalsemi.onewire.container.OneWireContainer;
import com.dalsemi.onewire.application.file.OWFileDescriptor;
import com.dalsemi.onewire.application.file.OWFileInputStream;
import com.dalsemi.onewire.application.file.OWFileOutputStream;
import com.dalsemi.onewire.application.file.OWFile;
import com.dalsemi.onewire.application.file.OWSyncFailedException;
import com.dalsemi.onewire.container.MemoryBank;
import com.dalsemi.onewire.container.PagedMemoryBank;

public class Jenny
{
  //start of class
public static void main(String args[]) {
	//print program information
	System.out.println("\nJenny 1.2 - kamikazi / blastzone.net\n");
	System.out.println("This software is licensed under the GNU GPL.");
	System.out.println("http://www.gnu.org/licenses/gpl.txt");
	iButtonIO iButtonSafe = new iButtonIO();
	int rescan = 1;
//this loop calls SelectiButton again if the user rescans the onewire bus
	do {
		//call  function to select the iButton
		rescan = SelectiButton(iButtonSafe);
	}while(rescan == 1);
//if we're here, we're done.
//call our destructors, if necessary
	iButtonSafe.finalize();
//end of main()
}


/**
 *  function SelectiButton
 *  Gets iButtons and allows user to select one to use
 *  Our program stack starts here
 *  Takes iButtonIO object
 *  Returns 1 to rescan the onewire bus, 0 to quit
*/
private static int SelectiButton(iButtonIO iButtonSafe) {
	int selection = -2;
	int return_value = 0;
	//scan for new iButtons
	//to change what types of buttons to scan, look at the GetiButtons function
	iButtonSafe.GetiButtons();
	Vector iButtonList = iButtonSafe.ListiButtons();
	//tests to see if there are any elegible iButtons
	if (iButtonList.size() == 0) {
		System.out.println("The program could not find any elegible iButtons.");
		System.out.println("Ensure the iButton is connected to your default adapter.");
	}
	//test to see if there is only one elegible iButton
	else if (iButtonList.size() == 1) {
		System.out.println("There is only one elegible iButton.");
		selection = 0;
	}
	//there is more than one elegible iButton
	else {
		//list valid iButtons
		for (int i = 0; i < iButtonList.size(); i++) {
			System.out.println("(" + i +") " + iButtonList.elementAt(i));
		}
	}

	//0 is now a special case. if selection is 0, there is only one ibutton
	//and the menu is unnecessary, so we filter it out.
	if (selection == 0)
		return_value = EnterPassword(iButtonSafe,0);
	else {
		//we're printing the menu here
		System.out.println("(r) Rescan for iButtons");
		System.out.println("(q) Quit");
		selection = EnterSelection(iButtonList.size(), false); //gets selection
		//we will either quit, rescan, or continue
		if (selection == -2)
			return_value = 1; //rescan
		else if (selection == -1)
			return_value = 0; //quit
		else
			return_value = EnterPassword(iButtonSafe,selection);
	}
	return return_value;
//end of function
}


/**
 *  function EnterPassword
 *  Obtains the password from the user.
 *  Takes iButtonIO object, integer of iButton to use
 *  Returns 1 to drop to 'select iButton' menu, 0 to exit program
*/
private static int EnterPassword(iButtonIO iButtonSafe,int selection) {
	int return_value;
	char[] password = null; //the password is stored in a char array
	System.out.print("Enter master password: ");
	System.out.flush();
	try {
		password = readPasswd(System.in, true);
	}
	catch (Exception e) {
		System.err.println("Error in Password: " + e.getMessage());
	}

	do {
		//if SelectFile returns <0, reprint the server list
		return_value = SelectFile(iButtonSafe, selection, password);
	}while(return_value < 0);

	//if we're here, we no longer need the password. destroy it.
	CharArrayZero(password);
	return return_value;
//end of function
}


/**
 *  function SelectFile
 *  Lists all files on the selected iButton, using the password to decrypt
 *  Takes iButtonIO object, integer of iButton to use, and the password
 *  Returns 1 to drop to 'select iButton' menu, 0 to exit program, -1 to re-list all files
*/
private static int SelectFile(iButtonIO iButtonSafe,int iButtonSafe_selection,char[] password) {
	int i;
	Vector iButtonFiles = new Vector(1);
	//populate the vector with all files
	try {
		iButtonFiles = iButtonSafe.ListFiles(iButtonSafe_selection);
	}
	catch (Exception e) {
		System.err.println("Error getting files: " +e.getMessage());
		return 1;
	}
	//remove files that do not start with "PASS"
	for(i = (iButtonFiles.size() -1); i >= 0; i--) {
		if(!((String)iButtonFiles.elementAt(i)).startsWith("PASS"))
			iButtonFiles.remove(i);
	}
	// our vector iButtonFiles contains only the PASS.* files.
	// now, we need to decrypt them and display the server and username.
	// depending on your machine and the number of files, this may take time.
	// we will also print a statusbar here.
	Vector iButtonFiles_decrypted = new Vector();
	int number_of_files = iButtonFiles.size();
	System.out.println("Decrypting " +number_of_files +" password files.");
	System.out.print("|");
	for(i = 0; i < number_of_files; i++) {
		System.out.print("-");
	}
	System.out.print("|\n ");
	for(i = 0; i < number_of_files; i++) {
		try {
			//read the file length and build a buffer
			iButtonSafe.GetReadFileStream(iButtonSafe_selection,(String)iButtonFiles.elementAt(i));
			byte[] readfile_buffer = new byte[iButtonSafe.ReadFileStreamLength()];
			iButtonSafe.CloseReadFileStream();

			//read the file into the buffer.
			iButtonSafe.GetReadFileStream(iButtonSafe_selection,(String)iButtonFiles.elementAt(i));
			iButtonSafe.ReadFileStream(readfile_buffer);
			iButtonSafe.CloseReadFileStream();

			//create an input stream from the buffer to make it easier to work with
			ByteArrayInputStream new_inputstream = new ByteArrayInputStream(readfile_buffer);

			//set up space to store the file data, and read it
			byte[] new_salt = new byte[8];
			byte[] new_count = new byte[4];
			byte[] new_ciphertext = new byte[readfile_buffer.length - 13];
			new_inputstream.read(new_salt,0,8);
			new_inputstream.read(new_count,0,4);
			new_inputstream.read(new_ciphertext,0,(readfile_buffer.length-13));

			//**Decryption**
			//the plaintext will come out as a byte array. change it to a
			//character array and add to the vector
			Integer new_count_as_Integer = new Integer(new String(new_count));
			EncryptedWrapper decrypt_engine = new EncryptedWrapper(new_count_as_Integer.intValue(),new_salt,new_ciphertext);
			char[] plaintext_data = ByteToCharArray(decrypt_engine.GetPlaintext(password));
			iButtonFiles_decrypted.addElement(plaintext_data);

			//we zero our byte arrays here, just to be on the safe side.
			//the Vector directly references the character arrays, so they are zeroed later.
			ByteArrayZero(new_count);
			ByteArrayZero(new_salt);
			ByteArrayZero(new_ciphertext);

			//finally, update our statusbar
			System.out.print("*");
		}
		catch (IOException e) {
			System.err.println("IO Exception in file PASS." +i +": " +e.getMessage());
		}
		catch (Exception e) {
			System.err.println("Unhandled Exception caught reading PASS." +i +" file: " + e.getMessage());
		}
	//end of for loop
	}
	//end our statusbar
	System.out.print("\n");

	//**MENU**
	//create our menu here
	int return_value = 0;
	//test to see if there are any files
	if (iButtonFiles_decrypted.size() == 0) {
		//there are no files. print error and drop to "select ibutton" menu.
		System.out.println("No password files were found.");
		return_value = 1;
	}
	//there are files
	else {
			System.out.println("          Server         Username");
			for (i = 0; i < iButtonFiles_decrypted.size(); i++) {
				char[] line = (char[])iButtonFiles_decrypted.elementAt(i);
				System.out.print("(" + i +") ");
				int k = 0;
				int tabs = 0;
				do {
					if(line[k] == '\t')
						tabs++;
					System.out.print(line[k]);
					k++;
				}while(tabs < 2);
				System.out.print("\n");
				/*  uncomment for pager
				if((i % 1) == 0) {
					//pause
					if(!PauseDisplay())
						i = iButtonFiles_decrypted.size() + 1;
				}
				*/
			}

	}

	//our menu includes options >= 0.
	//no need to give the user a chance to re-scan.
	System.out.println("(a) Add a new password");
	System.out.println("(r) Return to previous menu");
	System.out.println("(q) Quit");
	int selection = EnterSelection(iButtonFiles_decrypted.size(), true); //gets selection
	if (selection == -3) //create new
		return_value = AddEntry(iButtonFiles,iButtonSafe,iButtonSafe_selection,password );
	else if (selection == -2) //rescan the bus
		return_value = 1;
	else if (selection == -1) //quit
		return_value = 0;
	else
		return_value = ServerInfo(iButtonFiles_decrypted,iButtonFiles,selection,iButtonSafe,iButtonSafe_selection, password);

	//if we're here, we're done with this iButton.
	//destroy the character arrays.
	for(i=0; i < iButtonFiles_decrypted.size(); i++) {
		CharArrayZero((char[])iButtonFiles_decrypted.elementAt(i));
	}
	//System.out.println("Select Password returned: " +return_value);
	return return_value;
//end of function
}


/**
 *  function ServerInfo
 *  Displays the password for a server, offers ability to change or delete
 *  Takes vector of iButtonFile information, iButton file names, and the selected file locator.
 *  Also takes iButtonIO object, selection, and password char array to pass to files.
 *  Returns -1 to re-scan for files, 0 to quit the program
*/
private static int ServerInfo(Vector iButtonFiles_decrypted, Vector iButtonFiles, int selection,iButtonIO iButtonSafe, int iButtonSafe_selection, char[] password) {
	int return_value = -1; //by default anything that happens here will result in a re-scan.
	int menu_choice;
	//**Clear the screen here
	char[] server_data = (char[])iButtonFiles_decrypted.elementAt(selection);
	System.out.println("  Server    |    Username    |    Password    |    Date Set");
	//we use a for loop so we don't have to create a temporary string to display.
	for(int i=0; i < server_data.length; i++) {
		if(server_data[i] == '\t')
			System.out.print("  |  ");
		else
			System.out.print(server_data[i]);
	}
	System.out.print("\n");

	//print menu here
	System.out.println("(1) Change this entry");
	System.out.println("(2) Delete this entry");
	System.out.println("(r) Return to previous menu");
	System.out.println("(q) Quit");
	menu_choice = EnterSelection(3, false); //gets selection
	if (menu_choice == 1)
		ChangeEntry(iButtonFiles_decrypted, iButtonFiles, selection, iButtonSafe, iButtonSafe_selection, password);
	else if (menu_choice == 2)
		DeleteEntry(iButtonFiles, selection, iButtonSafe, iButtonSafe_selection);
	else if (menu_choice == -1)
		return_value = 0;

	//System.out.println("ServerInfo returned: " +return_value);
	return return_value;
//end of function
}


/**
 *  function ChangeEntry
 *  Provides an interface to change a password file entry
 *  Takes vector of iButtonFile information, iButton file names, selected file locator, iButtonIO, iButton selection, and password.
 *  Returns 1 on success, 0 if failure
*/
private static int ChangeEntry(Vector iButtonFiles_Decrypted, Vector iButtonFiles, int iButtonFiles_selection, iButtonIO iButtonSafe, int iButtonSafe_selection, char[] password) {
	int return_value = 0;
	//get the existing information and load it into a char array
	char[] server_data_buffer = (char[])iButtonFiles_Decrypted.elementAt(iButtonFiles_selection);
	//turn that array into a CharArrayReader to make access easier
	//also create a CharArrayWriter to help write the new file
	CharArrayReader server_data = new CharArrayReader(server_data_buffer);
	CharArrayWriter temp_space = new CharArrayWriter();
	char[] temp_character = new char[1];

	//get the server name
	do {
			try {
				server_data.read(temp_character,0,1);
				if(temp_character[0] != '\t') //we dont want any tabs.
					temp_space.write(temp_character,0,1);
			}
			catch (Exception e) {
				System.err.println("Exception caught changing data: " +e.getMessage());
				return 0;
			}
		}while(temp_character[0] != '\t');
		char[] server_name = temp_space.toCharArray();
		temp_space.reset();

	//get the user name
	do {
			try {
				server_data.read(temp_character,0,1);
				if(temp_character[0] != '\t') //we dont want any tabs.
					temp_space.write(temp_character,0,1);
			}
			catch (Exception e) {
				System.err.println("Exception caught changing data: " +e.getMessage());
				return 0;
			}
		}while(temp_character[0] != '\t');
		char[] user_name = temp_space.toCharArray();
	//we will force the user to change the password.
	//set up the date variable
	Date newdate = new Date();
	String date_set;
	char[] user_password = new char[1];
	char[] new_user_password = new char[1];

	//we have all our existing data. ask to change it.
	try {
		System.out.print("Enter server name [" +new String(server_name) +"]: ");
		char[] new_server_name = readPasswd(System.in, false);
		if(new_server_name == null)
			new_server_name = server_name;
		System.out.print("\nEnter user name [" +new String(user_name) +"]: ");
		char[] new_user_name = readPasswd(System.in, false);
		if(new_user_name == null)
			new_user_name = user_name;
		do {
			System.out.print("\nEnter password: ");
			user_password = readPasswd(System.in, true);
			System.out.print("\nEnter password again: ");
			new_user_password = readPasswd(System.in, true);
		}while(!CharArrayCompare(user_password,new_user_password));

		//set the current date/time
		date_set = newdate.toString();

		//write to our CharArrayWriter
		temp_space.reset();
		temp_space.write(new_server_name,0,new_server_name.length);
		temp_space.write("\t");
		temp_space.write(new_user_name,0,new_user_name.length);
		temp_space.write("\t");
		temp_space.write(user_password,0,user_password.length);
		temp_space.write("\t");
		temp_space.write(date_set);

		//**ENCRYPT**
		//we use a new salt/count.
		EncryptedWrapper new_encrypt = new EncryptedWrapper();
		Integer counter = new Integer(new_encrypt.SetCount(0));
		byte[] salt = new_encrypt.SetSalt(null);
		byte[] counterbytes = (counter.toString()).getBytes();
		byte[] created_ciphertext = new_encrypt.SetPlaintext(CharToByteArray(temp_space.toCharArray()),password);
		//we have all the information we need to write the file.
		iButtonSafe.GetWriteFileStream(iButtonSafe_selection,(String)iButtonFiles.elementAt(iButtonFiles_selection));
		iButtonSafe.WriteFileStream(salt);
		iButtonSafe.WriteFileStream(counterbytes);
		iButtonSafe.WriteFileStream(created_ciphertext);
		iButtonSafe.CloseWriteFileStream();
	}
	catch (Exception e) {
		System.err.println("Exception caught writing new data: " +e.getMessage());
		return 0;
	}
	return return_value;
//end of function
}


/**
 *  function DeleteEntry
 *  Deletes a file from the iButton
 *  Takes vector of iButton file names, the selected file locator, iButtonIO object and iButton selection int.
 *  Returns 1 on success, 0 if failure
*/
private static int DeleteEntry(Vector iButtonFiles, int iButtonFiles_selection, iButtonIO iButtonSafe, int iButtonSafe_selection) {
	int return_value = 0;
	System.out.print("\nAre you sure? (y/n): ");
	String tempstring = getString(1);
	tempstring.trim();
	if (tempstring.equalsIgnoreCase("y")) {
		if(iButtonSafe.DeleteFile(iButtonSafe_selection,(String)iButtonFiles.elementAt(iButtonFiles_selection)))
			return_value = 1;
	}
	else if (tempstring.equalsIgnoreCase("yes")) {
		if(iButtonSafe.DeleteFile(iButtonSafe_selection,(String)iButtonFiles.elementAt(iButtonFiles_selection)))
			return_value = 1;
	}

	return return_value;
//end of function
}


/**
 *  function Add Entry
 *  Files are stored using sequential file extensions.
 *  Looks for a missing file extension and fills it in. If fails, creates new file
 *  Takes vector of iButtonFile information, iButton file names, selected iButton, and the password
 *  Returns -1 on success (NOTE!), 0 if failure
*/
private static int AddEntry(Vector iButtonFiles, iButtonIO iButtonSafe, int iButtonSafe_selection, char[] password ) {
	int return_value = 0;
        //**GENERATE FILE CONTENT**
	//set up the date variable
	Date newdate = new Date();
	String date_set;
	//Ask for new data
	try {
		System.out.print("Enter server name: ");
		char[] new_server_name = readPasswd(System.in, false);
		System.out.print("\nEnter user name: ");
		char[] new_user_name = readPasswd(System.in, false);

		//the password arrays have to be setup outside of the do loop.
		char[] user_password = new char[1];
		char[] new_user_password = new char[1];
		do {
			System.out.print("\nEnter password: ");
			user_password = readPasswd(System.in, true);
			System.out.print("\nEnter password again: ");
			new_user_password = readPasswd(System.in, true);
		}while(!CharArrayCompare(user_password,new_user_password));

		//set the current date/time
		date_set = newdate.toString();

		//we need an array to feed the encryption.
		//so, we create a character array writer
		CharArrayWriter temp_space = new CharArrayWriter();
		temp_space.write(new_server_name,0,new_server_name.length);
		temp_space.write("\t");
		temp_space.write(new_user_name,0,new_user_name.length);
		temp_space.write("\t");
		temp_space.write(user_password,0,user_password.length);
		temp_space.write("\t");
		temp_space.write(date_set);

		//**ENCRYPT**
		//we use a new salt/count.
		EncryptedWrapper new_encrypt = new EncryptedWrapper();
		Integer counter = new Integer(new_encrypt.SetCount(0));
		byte[] salt = new_encrypt.SetSalt(null);
		byte[] counterbytes = (counter.toString()).getBytes();
		byte[] created_ciphertext = new_encrypt.SetPlaintext(CharToByteArray(temp_space.toCharArray()),password);
                //**END GENERATE FILE CONTENT**
		//we have all the information we need to write the file.
		//find an opening in the iButtonFiles list, if none, append to end.
		int extension = iButtonFiles.size();
		int i;
		for(i=0; i < iButtonFiles.size(); i++) {
			String temp_string = (String)iButtonFiles.elementAt(i);
			if(!temp_string.endsWith(ComposeString(i))) {
				extension = i;
				i = iButtonFiles.size() + 1;
			}
		}
		//write the file
		String filename = "PASS." + extension;
		iButtonSafe.GetWriteFileStream(iButtonSafe_selection,filename);
		iButtonSafe.WriteFileStream(salt);
		iButtonSafe.WriteFileStream(counterbytes);
		iButtonSafe.WriteFileStream(created_ciphertext);
		iButtonSafe.CloseWriteFileStream();
		return_value = -1;
	}
	catch (Exception e) {
		System.err.println("Exception caught writing new data: " +e.getMessage());
		return 0;
	}
	return return_value;
}


/**
 *  function PauseDisplay
 *  Pauses the display, gives the user the ability to stop displaying new entries
 *  Takes none
 *  Returns true to continue, false to stop
*/
private static boolean PauseDisplay() {
	boolean return_value = true;
	//System.out.println("Press any key to continue, Esc. to stop.");
	//InputStream temp_stream = new InputStream(System.in);
	//char temp_char = temp_stream.
	//System.err.println("The char was: " +temp_char);
	return return_value;

}

/**
 *  function CharArrayZero
 *  Zeros a character array, removing any sensitive information from memory
 *  Takes: character array
 *  Returns nothing
*/
private static void CharArrayZero(char[] input) {
	for(int i=0; i < input.length; i++) {
		input[i] = '0';
	}
}


/**
 *  function ByteArrayZero
 *  Zeros a byte array, removing any sensitive information from memory
 *  Takes: byte array
 *  Returns: nothing
*/
private static void ByteArrayZero(byte[] input) {
	for(int i=0; i < input.length; i++) {
		input[i] = (byte)0x00;
	}
}


/**
 *  function CharArrayCompare
 *  Compares two character arrays
 *  Takes 2 character arrays
 *  Returns false if different, true if the same
*/
private static boolean CharArrayCompare(char[] a,char[] b) {
	boolean same = true;
		if(a.length != b.length)
			same = false;
		else
			for(int i=0; i < a.length && same; i++) {
				if(a[i] != b[i])
					same = false;
			}
	return same;
}


/**
 *  function ComposeString
 *  Changes an integer into a string format for iButton extensions
 *  Takes integer
 *  Returns String
*/
private static String ComposeString(int i) {
	return new Integer(i).toString();
}


/**
 *  function CharToByteArray
 *  Changes a character array to a byte array using UTF-8 encoding
 *  Takes: char array
 *  Returns: UTF-8 encoded byte array
*/
private static byte[] CharToByteArray(char[] input) {
	// Create the encoder and decoder for UTF-8
	Charset charset = Charset.forName("UTF-8");
	CharsetEncoder encoder = charset.newEncoder();
	byte[] output_array = new byte[1];
    try {
	        // Convert a string to UTF-8 bytes in a ByteBuffer
	        // The new ByteBuffer is ready to be read.
	        ByteBuffer bbuf = encoder.encode(CharBuffer.wrap(input));
			output_array = bbuf.array();
		} catch (CharacterCodingException e) {
			System.err.println("A Coding exception was caught: " +e.getMessage());
			System.err.println("The program will continue, but your data may be damaged");
    	}
	return output_array;
}


/**
 *  function ByteToCharArray
 *  Changes a byte array to a character array using UTF-8 encoding
 *  Takes byte array
 *  Returns char array
*/
private static char[] ByteToCharArray(byte[] input) {
	// Create the encoder and decoder for UTF-8
	Charset charset = Charset.forName("UTF-8");
	CharsetDecoder decoder = charset.newDecoder();
	char[] output_array = new char[1];
    try {
	        // Convert a string to UFT-8 bytes in a ByteBuffer
	        // The new ByteBuffer is ready to be read.
	        CharBuffer cbuf = decoder.decode(ByteBuffer.wrap(input));
	        output_array = cbuf.array();
	    } catch (CharacterCodingException e) {
			System.err.println("A Coding exception was caught: " +e.getMessage());
			System.err.println("The program will continue, but your data may be damaged");
    	}
	return output_array;
}


/**
 *  Code used from Dallas Semiconductor iButton SDK
 * Reads string from user
 * returns int value from that string
 * used to determine menu considerations
 * returns -1 on error
*/
private static int EnterSelection(int max, boolean can_add) {
	int value = -999; //a marker value	
            do {
                System.out.print("Enter Selection: ");
                    try {
                    	String str = getString(1);
			str = str.trim();
			if (str.equalsIgnoreCase("q"))
				value = -1;
			else if (str.equalsIgnoreCase("r"))
				value = -2;
			else if (str.equalsIgnoreCase("a") && can_add)
				value = -3;
			else {
				value = Integer.parseInt(str);
				if(value >= max) {
					System.out.println("Invalid value, " + value +" must be less than " + max);
					value = EnterSelection(max, can_add);
				}
			}
                    }
                    catch (NumberFormatException e) {
			System.out.println("Invalid value: " +e.toString());
                        value = -999;
                    }
                }while(value == -999);
		return value;
}


/**
 *  Code used from Dallas Semiconductor iButton SDK
 * function getString
 * Retrieve user input from the console.
 * @param minLength minumum length of string
 * @return string entered from the console.
*/
private static String getString (int minLength) {
		  BufferedReader dis = new BufferedReader(new InputStreamReader(System.in));
	      String str;
	      boolean done = false;

	      try
	      {
	         do
	         {
	            str = dis.readLine();
	            if(str.length() < minLength)
	               System.out.print("String too short try again:");
	            else
	               done = true;
	         }
	         while(!done);

	         return str;
	      }
	      catch (java.io.IOException e)
	      {
	         System.out.println("Error in reading from console: " + e);
	      }
	      return "";
	}

 /**
  *  Code used from Dallas Semiconductor iButton SDK
  * Reads user password from given input stream.
  * obtained from java.sun.com
  * note that we return a char array, not a string.
  * this is intentional, for security!
 */
    private static char[] readPasswd(InputStream in, boolean hidden) throws IOException {
        PasswordView view = new PasswordView();

        char[] lineBuffer;
        char[] buf;
        int i;

        buf = lineBuffer = new char[128];

        int room = buf.length;
        int offset = 0;
        int c;
        if(hidden)
			view.start();

//start the loop, which unless broken, runs forever
loop:   while (true) {
            switch (c = in.read()) {
              case -1:
              case '\n':
                break loop;
				//break on newline

				// backspace
              case '\r':
                int c2 = in.read();
                if ((c2 != '\n') && (c2 != -1)) {
                    if (!(in instanceof PushbackInputStream)) {
                        in = new PushbackInputStream(in);
                    }
                    ((PushbackInputStream)in).unread(c2);
                } else
                    break loop;
              //are we out of room?
              default:
                if (--room < 0) {
                    buf = new char[offset + 128];
                    room = buf.length - offset - 1;
                    System.arraycopy(lineBuffer, 0, buf, 0, offset);
                    Arrays.fill(lineBuffer, ' ');
                    lineBuffer = buf;
                }
                buf[offset++] = (char) c;
                break;
            }
            System.out.print('\r');
        }

		//nothing was entered
        if (offset == 0) {
            return null;
        }
		//return the password
        char[] ret = new char[offset];
        System.arraycopy(buf, 0, ret, 0, offset);
        Arrays.fill(buf, ' ');
		if(hidden)
			view.halt();
        return ret;
    }

//end of class
}



/**
 * EncryptedWrapper class
 * contains all of the components required for the PKCS #5 encryption
 * can be constructed as empty, or with all fields filled
 * This class was developed to use des-ede3-cbc encryption
*/
class EncryptedWrapper {
    private int count = 0;
    private byte[] salt = {(byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00,(byte) 0x00};
    private byte[] ciphertext = null;

    /**
     * EncryptedWrapper constructor
     * just a null constructor
    */
    public EncryptedWrapper() {
	       //a null constructor
    }

    /**
     * EncryptedWrapper constructor
     * Takes: int count, byte array salt, byte array ciphertext
    */
    public EncryptedWrapper(int in_count, byte[] in_salt, byte[] in_ciphertext) {
		count = in_count;
		salt = in_salt;
		ciphertext = in_ciphertext;
	}


	//** Functions to return values **
	/**
	 * function GetSalt
	 * Takes nothing
	 * Returns byte array of the salt value
	*/
	public byte[] GetSalt() {
		return salt;
	}


	/**
	 * function GetCount
	 * Takes nothing
	 * Returns integer of the count value
	*/
	public int GetCount() {
		return count;
	}


	/**
	 * function GetCiphertext
	 * Takes nothing
	 * Returns byte array of the ciphertext value
	*/
	public byte[] GetCiphertext() {
		return ciphertext;
	}


	/**
	 * function GetPlaintext
	 * Takes char array of password
	 * Returns byte array of the plaintext value
	 * This function actually decrypts the value in ciphertext
	*/
	public byte[] GetPlaintext(char[] password) {
		//set up our keyfactory
		PBEKeySpec pbeKeySpec;
	    PBEParameterSpec pbeParamSpec;
    	SecretKeyFactory keyFac;

		byte[] plaintext = null;  //define memory space for our plaintext

	    // Create PBE parameter set
    	pbeParamSpec = new PBEParameterSpec(salt, count);

		try {
			pbeKeySpec = new PBEKeySpec(password);
	    	keyFac = SecretKeyFactory.getInstance("PBEWithMD5AndDES");
	    	SecretKey pbeKey = keyFac.generateSecret(pbeKeySpec);
	    	// Create PBE Cipher
	    	Cipher pbeCipher = Cipher.getInstance("PBEWithMD5AndDES");
	    	// Initialize PBE Cipher with key and parameters
	    	pbeCipher.init(Cipher.DECRYPT_MODE, pbeKey, pbeParamSpec);
			// Decrypt the ciphertext
	    	plaintext = pbeCipher.doFinal(ciphertext);
		}
		//The bad padding exception happens when we use the wrong password
		catch (BadPaddingException e) {
			System.err.println("This is not the password for this file.");
		}
		catch (Exception e) {
				System.err.println("Exception caught getting plaintext: " + e.getMessage());
		}
        return plaintext;
	}


	//** Functions to Set values **

	/**
	 * function SetCount
	 * Takes int count value
	 * Returns int count value
	 * To be PKCS#5 compliant, count must be > 1000. If count value is < 1000
	 * SetCount generates a new one between 1000 and 2000.
	*/
	public int SetCount(int in_count) {
		if(in_count < 1000)
		{
			Random rand = new Random();
			int sum = 0;
			for (int i=0; i<100; i++) {
      			sum += rand.nextInt(1000);
			}
			count = 1000 + sum/100;
		}
		else
		{
			count = in_count;
		}
		return count;
	}


	/**
	 * function SetSalt
	 * Takes byte array salt value
	 * Returns byte array salt value
	 * To be PKCS#5 compliant, salt must be at least 8 bytes.
	 * If passed a null value, generates a new salt
	 * Otherwise, sets salt value.
	*/
	public byte[] SetSalt(byte[] in_salt) {
		if (in_salt == null)
		{
			Random rand = new Random();
			int sum = 0;
			for (int j=0; j<8; j++) {
				for (int i=0; i<100; i++) {
      				sum += rand.nextInt(255);
				}
			salt[j] = ((byte) (sum/100));
			}
		}
		else
		{
			salt = in_salt;
		}
		return salt;
	}

	/**
	 * function SetCiphertext
	 * Takes byte array of ciphertext
	 * Returns byte array of ciphertext
	 * Use for pre-encrypted text
	*/
	public byte[] SetCiphertext(byte[] in_ciphertext) {
		ciphertext = in_ciphertext;
		return ciphertext;
	}

	/**
	 * function SetPlaintext
	 * Takes byte array to decrypt TO, char array password
	 * Returns byte array
	 * Decrypts ciphertext (use SetCiphertext) to the plaintext buffer passed.
	 * Returns byte array of decripted text
	*/
	public byte[] SetPlaintext(byte[] plaintext, char[] password) {
		//encrypt the plaintext using the salt and iteration counts
		//set up our Keyfactory
		PBEKeySpec pbeKeySpec;
	    PBEParameterSpec pbeParamSpec;
	   	SecretKeyFactory keyFac;

	    // Create PBE parameter set
	   	pbeParamSpec = new PBEParameterSpec(salt, count);

		try {
			pbeKeySpec = new PBEKeySpec(password);
	    	keyFac = SecretKeyFactory.getInstance("PBEWithMD5AndDES");
	    	SecretKey pbeKey = keyFac.generateSecret(pbeKeySpec);
	    	// Create PBE Cipher
	    	Cipher pbeCipher = Cipher.getInstance("PBEWithMD5AndDES");
	    	// Initialize PBE Cipher with key and parameters
	    	pbeCipher.init(Cipher.ENCRYPT_MODE, pbeKey, pbeParamSpec);
	    	// Encrypt the plaintext
	    	ciphertext = pbeCipher.doFinal(plaintext);
	   	}
		//just catch all exceptions
		catch (Exception e) {
			System.err.println("Exception caught generating ciphertext: " + e.getMessage());
		}
		return ciphertext;
	}


	/**
	 * default destructor
	 * The user is encouraged to call this directly to ensure plaintext is not left in memory
	*/
	protected void finalize() {
		count = 0;
		salt = null;
    	ciphertext = null;
	}

// end of class
}


/**
 * iButtonIO class
 * contains all of the components required for iButton IO
 * File access is fairly iButton-specific. Some casting may be required to refit for
 * standard filesystem access.
*/
class iButtonIO {
	//our class globals
	private DSPortAdapter adapter = null;		//the adapter
	private Vector owd_vect = new Vector(5);    //a list of OWContainers
	private OWFileInputStream owfis;			//stream to read from ibutton
	private OWFileOutputStream owfos;			//stream to write to ibutton


	/**
	 * default constructor
	 * obtains the iButton adapter and locks it
	*/
	public iButtonIO() {
		try {
			//let's find the adapter first
			adapter = OneWireAccessProvider.getDefaultAdapter();
			//lock the adapter
			adapter.beginExclusive(true);
		}
		catch (Exception e) {
			System.err.println("\nAn Exception was caught while initializing iButtonIO: " + e.getMessage());
		}
	}


	/**
	 * function GetiButtons
	 * scans for iButtons on the 1wire adapter, updates owd_vector with the device objects
	 * Takes nothing
	 * Returns vector size 0 if none, >0 if some.
	 * Use iButtonFamilies to determine which buttons to detect.
	*/
	public int GetiButtons() {
		OneWireContainer owd;
		//this family list is all NVRAM memory buttons
		//obtained from readme.txt in 1wire java API
		byte[] iButtonFamilies = { (byte) 0x04, (byte) 0x06, (byte) 0x08,(byte) 0x0A, (byte) 0x0C, (byte) 0x18, (byte) 0x1A, (byte) 0x1D };

		//we should first clear the owd_vector.
		owd_vect.removeAllElements();
		try {
		    // set some search restrictions
		    adapter.setSearchAllDevices();
		    adapter.targetFamily(iButtonFamilies);
		    adapter.setSpeed(DSPortAdapter.SPEED_REGULAR);

		    // enumerate through all the 1-Wire devices and collect them in a vector
		    for(Enumeration owd_enum = adapter.getAllDeviceContainers(); owd_enum.hasMoreElements(); )
		    {
		       owd = (OneWireContainer)owd_enum.nextElement();
		       owd_vect.addElement(owd);
		    }
		 }
		 catch(Exception e)
		 {
		    System.err.println("\nAn Exception was caught while getting iButtons: " + e.getMessage());
		    return 0;
		 }
		 return owd_vect.size();
	}


	/**
	 * function ListiButtons
	 * Takes nothing
	 * Returns a vector of iButton names, null if none.
	*/
	public Vector ListiButtons() {
		Vector iButtonList = new Vector(1);
		for (int i = 0; i < owd_vect.size(); i++) {
			iButtonList.addElement(((OneWireContainer)owd_vect.elementAt(i)).getName());
		}
		return iButtonList;
	}


	/**
	 * function ListFiles
	 * Takes int i, location in owd_vect of the iButton to scan
	 * Returns a vector of the file namess on the iButton root
	 * Also sorts list alphabetically
	*/
	public Vector ListFiles(int i) throws IOException{
		//to start in a different directory, place that inside the doublequotes
		OWFile owfile = new OWFile((OneWireContainer)owd_vect.elementAt(i),"");
		OWFile[] owfile_list;
		Vector output = new Vector(1);

		// get the list
		owfile_list = owfile.listFiles();
		// check for bad directory
		if (owfile_list != null) {
			//there is at least one file.
			//iterate and find the first file. Add it to the Vector.
			//once the first file is found, set first_file to true and continue.
			int k;
			boolean first_file = false;
			for(k = 0; k < owfile_list.length && !first_file; k++) {
				if(owfile_list[k].isFile()) {
					output.addElement(owfile_list[k].getName());
					first_file = true;
				}
			}

			//we have the first file in the array.
			//k is our current location in the owfile_list.
			//increment k by 1, because we already read that file.
			for (k = k; k < owfile_list.length; k++) {
				//iterate through all remaining files
				//reset first_file and reuse to let us know if we've placed the string
				first_file = false;
		        if (owfile_list[k].isFile())
		            {
		    			//we have a file. We need to place it in our Vector.
                                        //HOWEVER - we need to alphabetize by file name, then order numerically by extension.
                                        //get tempstring and break into name and extension
                                        //then, restore extension to ### format and recombine.
		    			String tempstring = new String(owfile_list[k].getName());
                                        String[] fileandext_array = tempstring.split("\\.");
                                      	 if(fileandext_array[1].length() == 1) {
                                            fileandext_array[1] = "00" +fileandext_array[1];
             				 }
             				 else if(fileandext_array[1].length() == 2) {
                 			    fileandext_array[1] = "0" +fileandext_array[1];
          				 }
               				tempstring = fileandext_array[0] +"." +fileandext_array[1];
                                        //end modifications.  tempstring is now in aaaa.nnn format
		    			for(int j = 0; j < output.size() && !first_file; j++) {
							if(tempstring.compareTo((String)output.elementAt(j)) < 0) {
								output.add(j,tempstring);
								first_file = true;
								//System.err.println("placed " +tempstring +" at " +j);
							}
						}
		    			if(!first_file) {
		    				//the file was never placed. place at the end.
		    				output.addElement(tempstring);
						}
				}
			owfile_list[k].close();
			}

		}
		// uncomment this section to print the file vector
		//System.err.println("Files from ListFiles");
		//for(int q = 0; q < output.size(); q++) {
		//	System.err.println(q +" " +(String)(output.elementAt(q)));
		//}
		//
		return output;
    }


	//** Functions to read a file from the iButton **
	/**
	 * function GetReadFileStream
	 * opens a new file stream for reading
	 * Takes integer location of the iButton in the owd_vect and a file name
	 * Returns nothing
	*/
	public void GetReadFileStream(int i, String file_name) {
		try {
			owfis = new OWFileInputStream(((OneWireContainer)owd_vect.elementAt(i)), file_name);
			}
		catch(Exception e)
			{
				System.err.println("\nAn Exception was caught while creating input stream: " + e.getMessage());
			}
	}


	/**
	 * function ReadFileStream
	 * opens a file stream for reading
	 * Takes byte array to use as file buffer
	 * Returns int of bytes read  -1 if EOF
	*/
	public int ReadFileStream(byte buf[]) throws IOException{
		return owfis.read(buf);
	}


	/**
	 * function ReadFileStreamLength
	 * Takes nothing
	 * Returns the length of an open filestream in integer.
	 * Because it's a filestream, after calling this function, close and reopen the stream
	 * This function leaves the pointer at EOF
	*/
	public int ReadFileStreamLength() throws IOException{
		int length = 0;
		byte[] block = new byte[1];
		do {
			length++;
		}while(owfis.read(block) != -1);
		return length;
	}


	/**
	 * function CloseReadFileStream
	 * closes a read file stream
	 * Takes nothing
     * Returns nothing
    */
	public void CloseReadFileStream() throws IOException {
		owfis.close();
	}
	//** End functions to read a file from the iButton **

	//** Functions to write a file to the iButton **
	/**
	 * function GetWriteFileStream
	 * constructs a file stream to write to the iButton
	 * Takes a OWContainer element, file name
	 * Returns nothing
	*/
	public void GetWriteFileStream(int i, String file_name) {
		try {
				owfos = new OWFileOutputStream(((OneWireContainer)owd_vect.elementAt(i)), file_name);
			}
		catch(Exception e)
			{
				System.err.println("\nAn Exception was caught while creating output stream: " + e.getMessage());
			}
	}


	/**
	 * function WriteFileStream
	 * writes a byte array to a filestream
	 * Takes byte array
	 * Returns nothing
	*/
	public void WriteFileStream(byte buf[]) throws IOException{
		owfos.write(buf);
	}


	/**
	 * function CloseWriteFileStream
	 * closes the file stream to a file
	 * closing is a critical step to writing!  You must call the close function to ensure integrity.
	 * Takes nothing.
	 * returns nothing. throws IOException
	*/
	public void CloseWriteFileStream() throws IOException {
		//get a file descriptor to flush to iButton
		OWFileDescriptor owfd = owfos.getFD();
		boolean done;
		// loop until sync is successful
		do {
			try {
		    	owfd.sync();
		        done = true;
		    	}
		    catch (OWSyncFailedException e)
		    	{
		      		System.err.println("An Error Occurred while Syncing to iButton.\n Please ensure iButton is in cradle");
		        	done = false;
		      	}
		   	} while (!done);
		owfos.close();
	}
	//** End functions to write a file to the iButton **


	/**
	 * function DeleteFile
	 * Function to delete a file from the iButton
	 * Takes integer of iButton location and file name string
	 * returns true on success, false on failure
	*/
	public boolean DeleteFile(int i, String file_name) {
		boolean return_value = true;
		try {
			OWFile owfile = new OWFile((OneWireContainer)owd_vect.elementAt(i), file_name);
			owfile.delete();

			OWFileDescriptor owfd = owfile.getFD();
			boolean done;
			// loop until sync is successful
			do {
				try {
					owfd.sync();
			        done = true;
		    	}
			    catch (OWSyncFailedException e)
		    	{
		      		System.err.println("An Error Occurred while Syncing to iButton.\n Please ensure iButton is in cradle");
		        	done = false;
		      	}
		   	} while (!done);
			owfile.close();
		}
		catch (Exception e) {
			System.err.println("An Error Occurred while deleting the file: " +e.getMessage());
			return_value = false;
		}
	return return_value;
	}


    /**
     * default destructor
     * releases the iButton adapter
    */
    protected void finalize() {
		try {
			//unlock the adapter
			adapter.endExclusive();
			//free the adapter
			adapter.freePort();
		}
		catch (Exception e) {
			System.err.println("\nAn Exception was caught while destroying iButtonIO: " + e.getMessage());
		}
	}

//end of class
}


/*
 * used from http://www.prabhums.org/weblogs/?postid=230
 * class to hide the password in the console
 */
class PasswordView extends Thread {
	private boolean typing = true;
	public void run() {
		while (typing) {
			System.out.print("\b "); // just put a backspace
		}
	}
	public synchronized void halt() {
		typing = false;
	}
}